// written by Zheng Gengbin
// irregular communication

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#if STDC_HEADERS
#include <stdlib.h>
#include <string.h>
#else
#include <strings.h>
extern "C" void exit(...);
#endif

#include <stdio.h>

#include "compilerApi.h"
#include "strategy.h"

#include "anls_att.h"


extern void change_array_reference(SgStatement * s, int counter, int nameflag, int indexflag, SgSymbol *stack=NULL,SgSymbol *stackBase=NULL);
extern void analyse(SgStatement * s);
extern void check_ghost(SgStatement * s);
extern void add_ghost(SgStatement * s);
extern void change_body(SgStatement * s,int);
extern void add_do_loop(SgStatement * s);


// called by trans_forall() in transforall.cc   (scatter)
// create an temp array , which is collapsed on every dim.
void create_lhs_section(SgStatement *s,SgArrayRefExp *array)
{
  Forall_header_def * fhdef = (Forall_header_def *)s->attributeValue(0, FHDER_Def);

  SgStatement * body = s->childList1(0);
  forall_reference * fref = (forall_reference *)body->attributeValue(0, Forall_Ref);

  SgSymbol * arraysymb  = array->symbol();
  SgArrayType *arraytype = isSgArrayType(arraysymb->type());

  int storage = getArrayStorage(arraysymb);

  SgSymbol * sddsymb=s->getSymbolInCurrentScope(mixstring, "dad", arraysymb->identifier());
  SgSymbol * grp = s->getSymbolInCurrentScope(mixstring, "grp", "root");
  SgExpression *grpExp = new SgVarRefExp(grp);
  SgStatement *new_array = make_new_array_data(s,sddsymb,arraysymb->type(),
						storage,-1,grpExp);
  s->insertStmtBefore(*new_array, *s->controlParent());

  for (int i=0; i<arraytype->dimension(); i++) {
	// pcrc_set_array_collapse
	SgStatement *coll_stmt = make_set_array_collapse(s,sddsymb,i+1,fhdef->getindextriple(i,0),fhdef->getindextriple(i,1));
  	s->insertStmtBefore(*coll_stmt, *s->controlParent());
  }
  // allocate memory;
  SgSymbol * mHandle;
  {
    SgSymbol * base = s->getSymbolInCurrentScope(mixstring, "tbs", forall_level, 0);

#ifdef MEM_ALLOC
    	mHandle = s->getSymbolInCurrentScope(mixstring, "mh", forall_level, 0);
	SgStatement *alloc=make_stack_alloc(s,arraysymb->type(),base,sddsymb,mHandle);
#else
	SgStatement *alloc=make_alloc(s,arraysymb->type(),base,sddsymb);
#endif

    s->insertStmtBefore(*alloc, *s->controlParent());
  }
  // reset base;
  {
    SgSymbol * basesym = s->getSymbolInCurrentScope(mixstring, "tbs", forall_level,0);

    SgStatement *funcall=make_reset_array_base(s,sddsymb,arraysymb->type(),basesym,mHandle);

    s->insertStmtBefore(*funcall, *s->controlParent());
  }
  // call delete_array;
  {
    SgStatement *delete_stmt = make_delete_array(s,sddsymb);
    s->insertStmtAfter(*delete_stmt, *s->controlParent());
  }
  // call real free;
  {
#ifdef MEM_ALLOC
    SgSymbol * base = s->getSymbolInCurrentScope(mixstring, "mh", forall_level, 0);
#else
    SgSymbol * base = s->getSymbolInCurrentScope(mixstring, "tbs", forall_level, 0);
#endif
    SgStatement* free_stmt = make_free(s,arraysymb->type(),base);
    s->insertStmtAfter(*free_stmt, *s->controlParent());
  }

  // copy to sdd0 
  {
  SgSymbol * sddsymb=s->getSymbolInCurrentScope(mixstring, "sdd", forall_level, 0);
  SgExpression *sddExp = new SgVarRefExp(sddsymb);
  SgSymbol * dadsymb=s->getSymbolInCurrentScope(mixstring, "dad", arraysymb->identifier());
  SgExpression *dadExp = new SgVarRefExp(dadsymb);
  SgAssignStmt *assign = new SgAssignStmt(*sddExp,*dadExp);
  s->insertStmtBefore(*assign, *s->controlParent());
  delete sddExp, dadExp, assign;
  }
}


// modify lhs array to a temp array in analyse phase
// called by analyse() in hf.cc
int regularize_lhs_array(SgStatement *s)
{
  int i;
  Forall_header_def * fhdef = (Forall_header_def *)s->attributeValue(0, FHDER_Def);

  SgStatement * body = s->childList1(0);
  forall_reference * fref = (forall_reference *)body->attributeValue(0, Forall_Ref);

  if (fref->checkRegular() == 1) return 0;
if (!silentFlag)
printf("Irregular forall reference - lhs!\n");

  SgArrayRefExp * array = (SgArrayRefExp *)fref->getexpression();  
  SgSymbol * arraysymb  = fref->getexpression()->symbol();
  SgType *basetype = arraysymb->type();
  ASSERT(basetype);
  if (basetype->hasBaseType()) basetype = basetype->baseType();
	
    // temp array's name "pc_a" if array "a"
  SgSymbol * tmpsymb=s->getSymbolInCurrentScope("pc", arraysymb->identifier(), NULL);
    // construct new array type
  SgArrayRefExp * newarray = new SgArrayRefExp(*tmpsymb);
  SgArrayType *arraytype = new SgArrayType(*basetype);
  arraytype->setDimension(fhdef->getindexnumber());
  for (i=0; i< fhdef->getindexnumber(); i++) {
	SgTripletExp *dim = new SgTripletExp(fhdef->getindextriple(i,0),fhdef->getindextriple(i,1),fhdef->getindextriple(i,2));
	arraytype->addRange(*dim);
  }
  tmpsymb->setType(*arraytype);
  newarray->setType(*basetype);

  for (i=0; i<fhdef->getindexnumber(); i++)
  {
	SgSymbol *indexsym = fhdef->getindexsymb(i);
	SgExpression *index = new SgVarRefExp(indexsym);
    	newarray->addSubscript(*index);
	delete index;
  }
 
  // save old array reference as attribute to array expression
  SgArrayRefExp *irr_array = isSgArrayRefExp(array->copyPtr());
  array->addAttribute(IRREG_ARRAY, irr_array , sizeof(irr_array));

  array->replace(*newarray);

  // analyse this new forall again, and delete the attributes.
  s->deleteAllAttribute(FHDER_Def);
  body->deleteAllAttribute(Forall_Ref);

  return 1;
}


void create_gather_tdd (SgStatement *s, int counter)
{
  int i;
  SgStatement * body = s->childList1(0);

  Forall_header_def * fhdef = (Forall_header_def *) s->getAttribute(0, FHDER_Def)->getAttributeData();

  forall_reference * lhsref = (forall_reference *)body->attributeValue(0, Forall_Ref);
  SgArrayRefExp * lhsarray = (SgArrayRefExp *) lhsref->getexpression();
  SgSymbol * lhsarraysymb  = lhsarray->symbol();

  forall_reference * rhsref = (forall_reference *)body->attributeValue(counter, Forall_Ref);
  SgArrayRefExp * rhsarray = (SgArrayRefExp *) rhsref->getexpression();
  SgSymbol * rhsarraysymb  = rhsarray->symbol();

//printf("lhsarraysymb:%s\n",lhsarraysymb->identifier());
//printf("rhsarraysymb:%s\n",rhsarraysymb->identifier());

  SgSymbol * tddsymb=s->getSymbolInCurrentScope(mixstring, "tdd", forall_level,counter);

  // new_array_copy
  {
  SgSymbol * sddsymb=s->getSymbolInCurrentScope(mixstring, "sdd", NULL,0);
  SgStatement *assign=make_new_array_copy(s,tddsymb,sddsymb);
  s->insertStmtBefore(*assign, *s->controlParent());
  }

  // allocate memory;
  SgSymbol * mHandle;
  {
    SgSymbol * base = s->getSymbolInCurrentScope(mixstring, "tbs", forall_level, counter);

#ifdef MEM_ALLOC
    	mHandle = s->getSymbolInCurrentScope(mixstring, "mh", forall_level, counter);
	SgStatement *alloc=make_stack_alloc(s,rhsarraysymb->type(),base,tddsymb,mHandle);
#else
	SgStatement *alloc=make_alloc(s,rhsarraysymb->type(),base,tddsymb);
#endif

    s->insertStmtBefore(*alloc, *s->controlParent());
  }
  // reset base;
  {
    SgSymbol * basesym = s->getSymbolInCurrentScope(mixstring, "tbs", forall_level,counter);

    SgStatement *funcall=make_reset_array_base(s,tddsymb,rhsarraysymb->type(),basesym,mHandle);

    s->insertStmtBefore(*funcall, *s->controlParent());
  }
  // call delete_array;
  {
    SgStatement *delete_stmt = make_delete_array(s,tddsymb);
    s->insertStmtAfter(*delete_stmt, *s->controlParent());
  }
  // call real free;
  {
#ifdef MEM_ALLOC
    SgSymbol * base = s->getSymbolInCurrentScope(mixstring, "mh", forall_level, counter);
#else
    SgSymbol * base = s->getSymbolInCurrentScope(mixstring, "tbs", forall_level, counter);
#endif
    SgStatement* free_stmt = make_free(s,rhsarraysymb->type(),base);
    s->insertStmtAfter(*free_stmt, *s->controlParent());
  }
}

// assign subscripts for subscript arrays for irregular array in lhs
void assign_lhs_subscripts(SgStatement *s,int counter,int dim,SgArrayRefExp *irr_array)
{
  SgSymbol *destdad=s->getSymbolInCurrentScope(mixstring, "subs", forall_level , counter,dim);

  Forall_header_def * fhdef = (Forall_header_def *) s->attributeValue(0, FHDER_Def);

  SgStatement * body = s->childList1(0);
  forall_reference * lhsref = (forall_reference *) (body->attributeValue(0, Forall_Ref));
  SgArrayRefExp *array = (SgArrayRefExp *)lhsref->getexpression();

    // create new assign expression
  SgValueExp *one = new SgValueExp(1);
  SgExpression *rhsval = &( irr_array->subscript(dim)->copy() - *one);
  SgExpression *lhsval = lhsref->getexpression()->copyPtr();
  SgAssignStmt * assign = new SgAssignStmt ( *lhsval, *rhsval);

  SgExpression *header = ((SgForallStmt *)s)->getHeader();

    // create a new forall statement
  SgForallStmt *newforall=new SgForallStmt(header->copy(),*assign);
  s->insertStmtAfter(*newforall, *s->controlParent());

    // recursive.
  inc_forall_level();

    // analyse this new forall statement , collected information
  analyse(newforall);

    // trans_forall
  check_ghost(newforall);	// maybe unknow - shift
  change_body(newforall,0);	// no need to create lhs section dad
  add_ghost(newforall);		// maybe unknow - shift

    // change lhs array reference to stack of sub arrays
  dec_forall_level();
#ifdef MEM_ALLOC
  SgSymbol * stack = s->getSymbolInCurrentScope(stackNameOfType(SgTypeInt()),"mb", NULL); 
#else
  SgSymbol * stack = s->getSymbolInCurrentScope(mixstring, stackNameOfType(SgTypeInt()), NULL); 
#endif
  SgSymbol * base = s->getSymbolInCurrentScope(mixstring, "tbs", forall_level, counter,dim);
  change_array_reference(newforall,0,1,1,stack,base);
  inc_forall_level();

    // preparation for OUT_OF_CORE
  if (is_out_of_core(array->symbol()))
  {
     forall_reference * fref = (forall_reference *) (newforall->childList1(0)->attributeValue(0, Forall_Ref));
     SgArrayRefExp *array = (SgArrayRefExp *)fref->getexpression();
     array->addAttribute(Outofcore_Def, destdad, sizeof(destdad));
  }
  add_do_loop(newforall);

    // decrease forall level.
  dec_forall_level();

  newforall->deleteStmt();
}


// assign subscripts for subscript arrays
void assign_rhs_subscripts(SgStatement *s,int counter,int dim)
{
  SgSymbol *destdad=s->getSymbolInCurrentScope(mixstring, "subs", forall_level , counter,dim);

  Forall_header_def * fhdef = (Forall_header_def *) s->attributeValue(0, FHDER_Def);

  SgStatement * body = s->childList1(0);
  forall_reference * fref = (forall_reference *) (body->attributeValue(counter, Forall_Ref));
  SgArrayRefExp *array = (SgArrayRefExp *)fref->getexpression();

  forall_reference * lhsref = (forall_reference *) (body->attributeValue(0, Forall_Ref));
  SgArrayRefExp *lhsarray = (SgArrayRefExp *)lhsref->getexpression();

  // create new assign expression
  SgValueExp *one = new SgValueExp(1);
  SgExpression *rhsval = &( array->subscript(dim)->copy() - *one);
  SgExpression *lhsval = lhsref->getexpression()->copyPtr();
  SgAssignStmt * assign = new SgAssignStmt ( *lhsval, *rhsval);

  SgExpression *header = ((SgForallStmt *)s)->getHeader();
  // create a new forall statement
  SgForallStmt *newforall=new SgForallStmt(header->copy(),*assign);
  s->insertStmtBefore(*newforall, *s->controlParent());

    // recursive. increase forall level
  inc_forall_level();
    // analyse this new forall statement , collected information
  analyse(newforall);

    // trans_forall
  check_ghost(newforall);	// maybe unknow - shift
  change_body(newforall,0);	// no need to create lhs section dad
  add_ghost(newforall);		// maybe unknow - shift

  dec_forall_level();

#ifdef MEM_ALLOC
  SgSymbol * stack = s->getSymbolInCurrentScope(stackNameOfType(SgTypeInt()),"mb", NULL); 
#else
  SgSymbol * stack = s->getSymbolInCurrentScope(mixstring, stackNameOfType(SgTypeInt()), NULL); 
#endif
  SgSymbol * base = s->getSymbolInCurrentScope(mixstring, "tbs", forall_level, counter,dim);
  change_array_reference(newforall,0,1,1,stack,base);
  inc_forall_level();

    // preparation for OUT_OF_CORE
  if (is_out_of_core(lhsarray->symbol()))
  {
     forall_reference * fref = (forall_reference *) (newforall->childList1(0)->attributeValue(0, Forall_Ref));
     SgArrayRefExp *array = (SgArrayRefExp *)fref->getexpression();
     array->addAttribute(Outofcore_Def, destdad, sizeof(destdad));
  }
  add_do_loop(newforall);

  dec_forall_level();

  newforall->deleteStmt();
}

// create temperary array to hold irregular subscripts
void create_subscripts (SgStatement *s , int counter,SgArrayRefExp *irr_array=NULL)
{
  Forall_header_def * fhdef = (Forall_header_def *) s->getAttribute(0, FHDER_Def)->getAttributeData();

  SgStatement * body = s->childList1(0);
  forall_reference * fref = (forall_reference *) (body->attributeValue(counter, Forall_Ref));
  SgArrayRefExp *array = (SgArrayRefExp *)fref->getexpression();

  int dim = array->numberOfSubscripts();
  if (irr_array) dim = irr_array->numberOfSubscripts();

  // deal all subscripts one by one
  for (int i=0; i< dim; i++)
  {
  	SgSymbol *destdad=s->getSymbolInCurrentScope(mixstring, "subs", forall_level , counter,i);
	SgSymbol *srcdad;
	if (counter==0)
  		srcdad=s->getSymbolInCurrentScope(mixstring, "dad", array->symbol()->identifier());
	else
  		srcdad=s->getSymbolInCurrentScope(mixstring, "tdd", forall_level , counter);
	// new array copy
	{
	SgStatement *assign=make_new_array_copy(s,destdad,srcdad);
        s->insertStmtBefore(*assign, *s->controlParent());
	}
  	// allocate memory;
	SgSymbol * mHandle;
  	{
    	SgSymbol * base = s->getSymbolInCurrentScope(mixstring, "tbs", forall_level, counter,i);
#ifdef MEM_ALLOC
    	mHandle = s->getSymbolInCurrentScope(mixstring, "mh", forall_level, counter,i);
	SgStatement *alloc=make_stack_alloc(s,SgTypeInt(),base,destdad,mHandle);
#else
	SgStatement *alloc=make_alloc(s,SgTypeInt(),base,destdad);
#endif
    	s->insertStmtBefore(*alloc, *s->controlParent());
  	}
  	// reset base;
  	{
    	SgSymbol * basesym = s->getSymbolInCurrentScope(mixstring, "tbs", forall_level,counter,i);
    	SgStatement *funcall=make_reset_array_base(s,destdad,SgTypeInt(),basesym, mHandle);
    	s->insertStmtBefore(*funcall, *s->controlParent());
  	}
	{
	// assign subscripts for subscript arrays
	if (counter == 0)
		assign_lhs_subscripts(s,counter,i,irr_array);
	else
		assign_rhs_subscripts(s,counter,i);
	}
  }
}

// delete subscripts for irregular in lhs.
void delete_lhs_subscripts (SgStatement *s , int counter)
{
  Forall_header_def * fhdef = (Forall_header_def *) s->getAttribute(0, FHDER_Def)->getAttributeData();

  SgStatement * body = s->childList1(0);
  forall_reference * fref = (forall_reference *) (body->attributeValue(counter, Forall_Ref));
  SgArrayRefExp *array = (SgArrayRefExp *)fref->getexpression();

  SgArrayRefExp *irr_array = (SgArrayRefExp *)array->attributeValue(0,IRREG_ARRAY);
  ASSERT(irr_array);

  for (int i= 0; i<= irr_array->numberOfSubscripts()-1; i++)
  {
  	SgSymbol *destdad=s->getSymbolInCurrentScope(mixstring, "subs", forall_level , counter,i);
  	SgSymbol *srcdad=s->getSymbolInCurrentScope(mixstring, "tdd", NULL , counter);

  	// call delete_array;
  	{
	SgStatement *delete_stmt = make_delete_array(s,destdad);
    	s->insertStmtAfter(*delete_stmt, *s->controlParent());
  	}
  	// call real free;
  	{
#ifdef MEM_ALLOC
    	SgSymbol * mHandle = s->getSymbolInCurrentScope(mixstring, "mh", forall_level, counter,i);
#else
    	SgSymbol * mHandle = s->getSymbolInCurrentScope(mixstring, "tbs", forall_level, counter,i);
#endif
    	SgStatement *free_stmt = make_free(s,array->type(),mHandle);
 	s->insertStmtAfter(*free_stmt, *s->controlParent());
 	}
  }
}

// delete subscripts for irregular in rhs.
void delete_rhs_subscripts (SgStatement *s , int counter)
{
  Forall_header_def * fhdef = (Forall_header_def *) s->getAttribute(0, FHDER_Def)->getAttributeData();

  SgStatement * body = s->childList1(0);
  forall_reference * fref = (forall_reference *) (body->attributeValue(counter, Forall_Ref));
  SgArrayRefExp *array = (SgArrayRefExp *)fref->getexpression();

  for (int i= array->numberOfSubscripts()-1; i>=0; i--)
  {
  	SgSymbol *destdad=s->getSymbolInCurrentScope(mixstring, "subs", forall_level , counter,i);
  	SgSymbol *srcdad=s->getSymbolInCurrentScope(mixstring, "tdd", NULL , counter);

  	// call delete_array;
  	{
	SgStatement *delete_stmt = make_delete_array(s,destdad);
    	s->insertStmtBefore(*delete_stmt, *s->controlParent());
  	}
  	// call real free;
  	{
#ifdef MEM_ALLOC
    	SgSymbol * mHandle = s->getSymbolInCurrentScope(mixstring, "mh", forall_level, counter,i);
#else
    	SgSymbol * mHandle = s->getSymbolInCurrentScope(mixstring, "tbs", forall_level, counter,i);
#endif
    	SgStatement *free_stmt = make_free(s,array->type(),mHandle);
 	s->insertStmtBefore(*free_stmt, *s->controlParent());
 	}
  }
}
